#!/usr/bin/env python

"""
Extract an unweighted grammar from an XML file in the SMULTRON treebank
format.

With the "printtrees" option passed, instead of spitting out a grammar, spit
out bracketed parse trees for every sentence.

Based, originally, on an example by Dave Kuhlman from
http://www.rexx.com/~dkuhlman/pyxmlfaq.html, but thoroughly hacked up.
"""

import sys, string
from xml.sax import handler, make_parser

import nltk
from nltk.grammar import ContextFreeGrammar
from nltk.grammar import Production
from nltk.grammar import Nonterminal
from nltk.tree import Tree

from util import replacepunctuation
from util import production_sort

MAX_SENTENCE_LEN = 15

class GrammarExtractor(handler.ContentHandler):
    def __init__(self, outfile):
        self.outfile = outfile
        self.productions = set()
        self.sentencetrees = []
        self.cur_nonterm = None

        # keys are node IDs, values are Nonterminal objects
        self.ids_to_cats = {}
        self.ids_to_spans = {}
        self.ids_to_trees = {}

        # position in the sentence
        self.cur_index = 0
        self.cur_id = None
        self.cur_lower = None
        self.cur_upper = None
        self.cur_targets = []
        self.cur_graphroot = None

    def startElement(self, name, attrs):
        if name == "s":
            ## starting a new sentence.
            self.ids_to_cats = {}
            self.ids_to_spans = {}
            self.ids_to_trees = {}
            self.cur_index = 0

        if name == "graph":
            self.cur_graphroot = str(attrs.get("root"))

        if name == "nt":
            ## When we see a new nonterminal, start tracking which edges attach
            ## to it.
            nodeid = str(attrs.get("id"))
            self.cur_id = nodeid
            cat = str(attrs.get("cat"))
            cat = replacepunctuation(cat)
            self.cur_nonterm = Nonterminal(cat)
            self.cur_targets = []

            self.ids_to_cats[nodeid] = Nonterminal(cat)
            self.cur_lower = None
            self.cur_upper = None

        if name == "edge":
            ## skip sentences longer than 15 words.
            if self.cur_index > MAX_SENTENCE_LEN: return

            ## add an edge. When we're done with the current nt, add a new
            ## production to the production list.
            targetid = str(attrs.get("idref"))
            targettype = self.ids_to_cats[targetid]
            self.cur_targets.append(targetid)

            ## when you see a new edge, update the current node's lower and
            ## upper bounds appropriately.
            if targetid not in self.ids_to_spans:
                print self.ids_to_spans, targetid
                assert False

            edge_lower, edge_upper = self.ids_to_spans[targetid]

            if self.cur_lower is None:
                self.cur_lower = edge_lower
            else:
                self.cur_lower = min(edge_lower, self.cur_lower)
            if self.cur_upper is None:
                self.cur_upper = edge_upper
            else:
                self.cur_upper = min(edge_upper, self.cur_upper)

        if name == "t":
            ## This is where we build up the lexicon.
            word = str(attrs.get("word"))
            pos = replacepunctuation(str(attrs.get("pos")))
            nodeid = str(attrs.get("id"))

            self.ids_to_cats[nodeid] = pos
            self.ids_to_spans[nodeid] = (self.cur_index, self.cur_index)
            self.ids_to_trees[nodeid] = pos
            self.cur_index += 1

    def endElement(self, name):
        if name == "nt":
            ## skip sentences longer than 15 words.
            if self.cur_index > MAX_SENTENCE_LEN: return

            targets = self.cur_targets
            lowers = [self.ids_to_spans[nodeid][0] for nodeid in targets]

            pairs = zip(targets, lowers)
            pairs.sort(key=lambda p: p[1])
            sortedtargets = [pair[0] for pair in pairs]

            cats = [self.ids_to_cats[nodeid] for nodeid in sortedtargets]
            prod = Production(self.cur_nonterm, cats)
            self.productions.add(prod)

            self.ids_to_spans[self.cur_id] = (self.cur_lower, self.cur_upper)


            children = [self.ids_to_trees[nodeid] for nodeid in sortedtargets]
            self.ids_to_trees[self.cur_id] = Tree(str(self.cur_nonterm),
                                                  children)

        if name == "graph":
            ## skip sentences longer than 15 words.
            if self.cur_index > MAX_SENTENCE_LEN: return

            lhs = Nonterminal("TOP")
            rhs = [self.ids_to_cats[self.cur_graphroot]]
            prod = Production(lhs, rhs)
            self.productions.add(prod)

            self.sentencetrees.append(self.ids_to_trees[self.cur_graphroot])

def extract(inFileName, printtrees=False):
    outFile = sys.stdout
    handler = GrammarExtractor(outFile)
    parser = make_parser()
    parser.setContentHandler(handler)
    parser.parse(inFileName)

    if not printtrees:
        productions = list(handler.productions)
        productions.sort(cmp=production_sort)
        start = Nonterminal("TOP")
        grammar = ContextFreeGrammar(start, productions)

        theproductions = grammar.productions()
        for prod in theproductions:
            print prod
    else:
        trees = handler.sentencetrees
        for tree in trees:
            print tree.pprint(margin=sys.maxint)

def main():
    args = sys.argv[1:]
    if len(args) not in [1,2]:
        print 'usage: python extractgrammar.py infile.xml [printtrees]'
        sys.exit(-1)

    printtrees = False
    if len(args) == 2:
        printtrees = (args[1] == "printtrees")
    extract(args[0], printtrees)

if __name__ == '__main__':
    main()
